@@ -2,8 +2,8 @@
 # archiver.py
 #
 
-import os, sys, socket, re, time, smtplib, calendar, StringIO, logging, logging.handlers, locale, gzip
-import urllib, urllib2, urlparse
+import os, sys, socket, re, time, smtplib, calendar, StringIO, logging, logging.handlers, locale, gzip, posixpath
+import urllib, urllib2, urlparse, ftplib
 from datetime import datetime, timedelta
 import email
 from email.Utils import parseaddr, formataddr, formatdate
@@ -28,7 +28,7 @@ domain_nameRE=re.compile('^'+dot_atom+'$')
 email_addressRE=re.compile('^'+postfix_restricted_rfc2822_email_address+'$')
 valid_ipRE=re.compile('^([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$')
 valid_hostnameRE=re.compile('^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\\-]*[a-zA-Z0-9])\\.)*([A-Za-z]|[A-Za-z][A-Za-z0-9\\-]*[A-Za-z0-9])$')
-ftp_url_re=re.compile('ftp://(?:(?P<username>[\w\.\-\+]+)(?::(?P<password>[\w\.\-\+]+))?@)?(?P<host>[a-zA-Z0-9\.\-]+)(?::(?P<port>\d+))?(?:(?P<directory>/.*))?')
+ftp_url_re=re.compile(r'ftp://(?:(?P<username>[\w\.\-\+%!$&\'\(\)*,;=#]+)(?::(?P<password>[\w\.\-\+%!$&\'\(\)*,;=#]+))?@)?(?P<host>[a-zA-Z0-9\.\-]+)(?::(?P<port>\d+))?(?:(?P<directory>/.*))?')
 
 # from FormEncode
 url_re = re.compile(r'''
@@ -190,6 +190,34 @@ def socket_readlines(chan):
     if line_buf:
         yield line_buf
 
+# ---------------------------------------------------------------------------
+def ftp_pre_create_dir(host, port, username, password, directory):
+    ftpcli=ftplib.FTP()
+    try:
+        ftpcli.connect(host, port, timeout=10)
+    except TypeError:
+        # python 2.5 and below have no timeout
+        ftpcli=ftplib.FTP()
+        ftpcli.connect(host, port)
+           
+    ftpcli.login(username, password)
+    dir=directory
+    dirs=[]
+    while dir and dir!='/':
+        dirs.insert(0, dir)
+        dir=posixpath.dirname(dir)
+
+    for dir in dirs:
+        try:
+            ftpcli.mkd(dir)
+        except:
+            pass
+    try:
+        ftpcli.quit()
+    except:
+        pass
+    ftpcli.close()
+
 # ---------------------------------------------------------------------------
 def _list_dir(target_dir, log):
     total_size=0
