@@ -117,6 +117,18 @@ class FTPsession(SocketServer.BaseRequestHandler):
     def cmd_cdup(self, args):
         self.response("553 already at the root dir")
     
+    def cmd_mkd(self, args):
+        target=os.path.join(self.root, args[0])
+        if not os.path.isdir(target):
+            try:
+                os.mkdir(target)
+            except OSError, e:
+                self.response('521 "%s" %s' % (args[0], e,))
+            else:
+                self.response('257 "%s" directory created' % (args[0], ))
+        else:
+            self.response('521 "%s" directory already exists' % (args[0], ))
+            
     def cmd_retr(self, args):
         if not args:
             self.response("500 retr syntax error")
