@@ -55,8 +55,7 @@ class MySCPClient(scp.SCPClient):
             chan.sendall('\x00')
 
     # -----------------------------------------------------------------------
-    def get(self, remote_path, local_path = '',
-            recursive = False, preserve_times = False):
+    def get(self, remote_path, local_path = '', recursive = False, preserve_times = False):
         self._recv_topdir = local_path or os.getcwd()
         return scp.SCPClient.get(self, remote_path, local_path, recursive, preserve_times)
     
@@ -70,6 +69,7 @@ class MySCPClient(scp.SCPClient):
             self.channel.send('\x01')
             raise scp.SCPException('Bad directory format')
         else:
+            print 'ASX', self._recv_dir, self._recv_topdir
             if self._recv_dir==self._recv_topdir and (not os.path.exists(self._recv_dir) or os.path.isdir(self._recv_dir)):
                 # use self._recv_dir instead of os.path.join(self._recv_dir, parts[2])
                 path=self._recv_dir
@@ -143,6 +143,7 @@ class Ghettovcb(Archiver):
     initiate_backup_re=re.compile(r'.*info: Initiate backup for (?P<name>.+)')
     backup_duration_re=re.compile(r'.*info: Backup Duration:')
     config_vm_backup_volume_re=re.compile(r'.*info:\s*CONFIG\s*-\s*VM_BACKUP_VOLUME\s*=\s*(?P<vm_backup_volume>.+)')
+    config_disk_backup_format_re=re.compile(r'.*info:\s*CONFIG\s*-\s*DISK_BACKUP_FORMAT\s*=\s*(?P<disk_backup_format>.+)')
     config_vm_backup_dir_naming_convention_re=re.compile(r'.*info:\s*CONFIG\s*-\s*VM_BACKUP_DIR_NAMING_CONVENTION\s*=\s*(?P<vm_backup_dir_naming_convention>.+)')
     find_vmdk_re=re.compile(r'.*debug: findVMDK\(\) - Searching for VMDK: "(?P<vmdk>.+)" to backup')
     ftp_url_re=ftp_url_re
@@ -179,6 +180,7 @@ class Ghettovcb(Archiver):
         self.upload_thread=None
         self.upload_queue=Queue.Queue()
         self.uploaded_filenames=dict()
+        self.download_script=None
         self.ftp_local=None
         self.ftp_local_default=None
         self.ftp_server=None
@@ -404,7 +406,7 @@ class Ghettovcb(Archiver):
         return errors, warnings, extra
 
     # -----------------------------------------------------------------------
-    def upload_queue_add(self, protocol, con, vm, source, destination):
+    def upload_queue_add_old(self, protocol, con, vm, source, destination):
         self.upload_queue.put((protocol, con, vm, source, destination))
         if not self.upload_thread:
             self.upload_thread=threading.Thread(target=self.uploader_thread, name='uploader')
@@ -412,7 +414,7 @@ class Ghettovcb(Archiver):
             self.upload_thread.start()
         
     # -----------------------------------------------------------------------
-    def uploader_thread(self):
+    def uploader_thread_old(self):
         current_vm=None
         self.upload_status=''
         if self.upload_status_code==None:
@@ -520,7 +522,6 @@ class Ghettovcb(Archiver):
                 root=''
             ftp['directory']=root
         
-        
         try:
             if ftp['directory']:
                 os.makedirs(ftp['directory'])
@@ -542,99 +543,291 @@ class Ghettovcb(Archiver):
         return True
          
     # -----------------------------------------------------------------------
-    def upload_vmdk(self, vm, root_backup_dir, vmdk, now):
-        type, target=self.destination.match(now, self.night_shift, variables=dict(vm=vm))
-        # print "ASX upload_vmdk vm=%s type=%s target=%s root_backup_dir=%s vmdk=%s" %(vm, type, target, root_backup_dir, vmdk)
-        if target and type in ('move', 'copy'):
+    def upload_vm(self, current_vm, root_backup_dir, now, compressed):
+        if not self.upload_thread:
+            self.upload_thread=threading.Thread(target=self.uploader_thread, name='uploader')
+            self.upload_thread.daemon=True
+            self.upload_thread.start()
+        
+        self.upload_queue.put((current_vm, root_backup_dir, now, compressed))
+
+    # -----------------------------------------------------------------------
+    def uploader_thread(self):
+        try:
+            if not self._uploader_thread():
+                self.upload_status_code='ERR'
+        except Exception, e:
+            self.upload_status_code='ERR'
+            self.log.exception('unexpected error in uploader_thread: %s', e)
+            
+    # -----------------------------------------------------------------------
+    def _uploader_thread(self):
+
+        self.upload_status=''
+        self.download_script=''
+        
+        if self.upload_status_code==None:
+            self.upload_status_code='OK'
+        while True:
+            current_vm, root_backup_dir, now, compressed=self.upload_queue.get(True)
+            
+            if current_vm==None:
+                break
+
+            type, target=self.destination.match(now, self.night_shift, variables=dict(vm=current_vm))
+            # print "ASX upload_vmdk vm=%s type=%s target=%s root_backup_dir=%s vmdk=%s" %(current_vm, type, target, root_backup_dir, vmdk)
+            if not target or type not in ('move', 'copy'):
+                continue
+                
             ftp_match=self.ftp_url_re.match(target)
 
             if ftp_match:
                 ftp=ftp_match.groupdict()
                 if ftp['directory']==None:
                     ftp['directory']=''
+                if ftp['port']!=None:
+                    ftp['port']=int(ftp['port'])
                     
                 if ftp['host']=='localhost':
                     if not self.ftp_server:
                         try:
                             res=self.start_ftp_server()
                         except Exception, e:
-                            self.log.exception('starting ftp server')
-                            return
+                            self.log.exception('cannot start built-in ftp server')
+                            return False
                         else:
-                            if res:
-                                self.log.info('ftpserver listening on %s:%d', self.ftp_server.socket.getsockname()[0], self.ftp_server.socket.getsockname()[1])
-                            else:
-                                return
+                            if not res:
+                                return False
+                            self.log.info('ftpserver listening on %s:%d', self.ftp_server.socket.getsockname()[0], self.ftp_server.socket.getsockname()[1])
+
                     ftp['host']=self.ftp_local_default['host']
                     for k in ('port', 'username', 'password'):
                         if ftp.get(k)==None:
                             ftp[k]=self.ftp_local_default[k]
-                
                 else:
+                    if not ftp['port']:
+                        ftp['port']=21
+                    if not ftp['username']:
+                        ftp['username']='anonymous'
+                    if not ftp['password']:
+                        ftp['password']='anonymous@'
+                        
                     # try to create directories on the remote ftp server from 
                     # this host. If it fail the directory must already exist.   
                     try:
-                        ftpcli=ftplib.FTP(ftp['host'], ftp['username'], ftp['password'], timeout=10)
                         dir=ftp['directory']
-                        dirs=[]
-                        while dir and dir!='/':
-                            dirs.insert(0, dir)
+                        if compressed:
                             dir=posixpath.dirname(dir)
-                            print dir
-
-                        for dir in dirs:
-                            try:
-                                ftpcli.mkd(dir)
-                            except:
-                                pass
-                        try:
-                            ftpcli.quit()
-                        except:
-                            pass
-                        ftpcli.close()
-                    except:
+                        ftp_pre_create_dir(ftp['host'], ftp['port'], ftp['username'], ftp['password'], dir)
+                    except Exception, e:
+                        self.log.exception('creating directory: %s', dir)
                         pass
-                            
-                                
+                    else:
+                        self.uploaded_filenames[dir]=None
+                        
                 target=ftp['directory']
                 protocol, con='ftp', ftp
+
+                if not self.download_script:
+                    self.download_script+='#\n'
+                    self.download_script+='# script generated by MKSBackup to download backup back to ESX(i) host\n'
+                    self.download_script+='#\n'
+                    self.download_script+='HOST="%s"\n' % (con['host'], )
+                    self.download_script+='PORT=%d\n' % (con['port'], )
+                    self.download_script+='USERNAME="%s"\n' % (con['username'], )
+                    self.download_script+='PASSWORD="your_password"\n'
+                    if compressed:
+                        self.download_script+='ROOT_DIR="%s"\n' % (posixpath.dirname(root_backup_dir), )
+                    else:
+                        self.download_script+='ROOT_DIR="%s"\n' % (posixpath.dirname(posixpath.dirname(root_backup_dir)), )
+                    
+                self.download_script+='#\n'
+                self.download_script+='# Download of VM: %s\n' % (current_vm, )
+                self.download_script+='#\n'
+                self.download_script+='DOWNLOAD_DIR="${ROOT_DIR}/download-%s"\n' % (current_vm, )
+                self.download_script+='mkdir -p "${DOWNLOAD_DIR}"\n'
                 
             else:
-                # create the target dir, if don't exist
+                # SSH
+                # create the target directory, if don't exist
                 try:
-                    if not os.path.isdir(target):
-                        os.makedirs(target)
+                    if compressed:
+                        dir=os.path.dirname(target)
+                    else:
+                        dir=target
+                    if not os.path.isdir(dir):
+                        os.makedirs(dir)
                 except Exception, e:
-                    self.log.error('creating directory %s: %s', target, e)
-                    return
+                    self.log.error('creating directory %s: %s', dir, e)
+                    return False
 
                 protocol, con='ssh', self.t
-
-            # search for files older than vmdk
-            if vmdk:
-                # file older then vmdk
-                find_cmd='find "%s" -type f \! -newer "%s"' % (root_backup_dir, posixpath.join(root_backup_dir, vmdk))
-            else:
-                # all files in the directory
-                find_cmd='find "%s" -type f' % (root_backup_dir, )
                 
-            exit_code, output=RunCommand(self.t, find_cmd, log=None)
-            # print 'ASX\n', output
-            if exit_code:
-                self.log.error('cannot list %s: %r',root_backup_dir, output)
-                self.upload_status_code='ERR'                
+            if compressed:
+                find_src=posixpath.join(root_backup_dir, compressed)
             else:
-                for filename in output.split('\n'):
-                    filename=filename.strip()
-                    if filename and not filename.startswith('.lck-') and filename not in self.uploaded_filenames:
-                        self.uploaded_filenames[filename]=None
-                        relfilename=filename[len(root_backup_dir):]
+                find_src=root_backup_dir
+
+            find_cmd='find "%s" -type f -exec stat -c "%%s:%%Y:%%n" {} \;' % (find_src, )
+
+            start=time.time()
+            status='OK'
+            self.upload_status+='upload_start-%s=%s\r\n' % (current_vm, time.ctime(start))
+            self.upload_status+='upload_start_epoch-%s=%d\r\n' % (current_vm, start)
+            self.upload_status+='upload_src-%s=%s\r\n' % (current_vm, find_src)
+            self.upload_status+='upload_dst-%s=%s\r\n' % (current_vm, target)
+            
+            filenames=set()
+            finished=False
+            while not finished:
+                # get epoch from ESX(i)
+                time.sleep(10)
+                exit_code, output=RunCommand(self.t, 'date +%s', log=None)
+                if exit_code:
+                    self.log.error('cannot get epoch: %r', output)
+                    return False
+                
+                finished=current_vm in self.finished_vm
+                # print 'ASX finished', current_vm, finished
+                esx_epoch=int(output.split('\n')[0])
+                exit_code, output=RunCommand(self.t, find_cmd, log=None)
+                if exit_code:
+                    self.log.error('cannot get file list of %s: %r', find_src, output)
+                    return False
+
+#      2496:1302887189:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01.vmx
+#1190330368:1302887225:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/4970f796-06563af5-553e-000e0c7 7dba/cos5-tiny 1-s001.vmdk
+#     65536:1302887191:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/4970f796-06563af5-553e-000e0c7 7dba/cos5-tiny 1-s002.vmdk
+#       528:1302887225:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/4970f796-06563af5-553e-000e0c7 7dba/cos5-tiny 1.vmdk
+#2072182784:1302887332:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s001.vmdk
+#2102722560:1302887444:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s002.vmdk
+#2106982400:1302887560:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s003.vmdk
+#2102001664:1302887678:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s004.vmdk
+#   4259840:1302887678:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01-s005.vmdk
+#       649:1302887678:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/eg01.vmdk
+#        30:1302887678:/vmfs/volumes/SATA500-0/backup/eg01/eg01-2011-04-15_17-05-56/STATUS.ok
+
+                files=dict()
+                for line in output.split('\n'):
+                    if not line:
+                        continue
+                    size, epoch, filename=line.split(':', 2)
+                    if filename.startswith('.lck-'):
+                        continue
+                    files[filename]=(int(size), int(epoch), line)
+                         
+                for filename, (size, epoch, line) in files.iteritems():
+                    if line in self.uploaded_filenames:
+                        continue
+                    if not finished:
+                        size, epoch=int(size), int(epoch)
+                        if epoch+60>esx_epoch and current_vm not in self.finished_vm:
+                            # file last modified < 30s ago, skip 
+                            continue
+
+                        match=re.match('.*-s(?P<number>\d{3}).vmdk$', filename)
+                        if match:
+                            number=int(match.group('number'))
+                            previous_part=filename[:match.start('number')]+'%03d' % (number-1, )+filename[match.end('number'):]
+                            if previous_part in files and files[previous_part][2] not in self.uploaded_filenames:
+                                # part of 2gbsparse but previous part not yet uploaded
+                                continue
+
+                    source=filename
+                                
+                    self.uploaded_filenames[line]=None
+                    if compressed:
+                        relfilename='%s.tar.gz' % (current_vm, )
+                    else:
+                        relfilename=source[len(root_backup_dir):]
                         relfilename=relfilename.lstrip('/')
-                        if protocol=='ftp':
-                            destination=posixpath.join(target, relfilename)
+                    
+                    destination=posixpath.join(target, relfilename)
+                    dest_dir=posixpath.dirname(destination)
+                    if protocol=='ftp':
+                            
+                        if dest_dir not in self.uploaded_filenames:
+                            self.download_script+='mkdir -p "${DOWNLOAD_DIR}/%s"\n' % (posixpath.dirname(relfilename), )
+                            try:
+                                ftp_pre_create_dir(ftp['host'], ftp['port'], ftp['username'], ftp['password'], dest_dir)
+                            except Exception, e:
+                                self.log.error('cannot create directory, try to create it yourself on ftp server: %s', dest_dir)
+                            else:
+                                self.uploaded_filenames[dest_dir]=None
+                            
+                        ftp_line='ftpput -P %s -u "%s" -p "%s" %s "%s" "%s"' % (con['port'], con['username'], con['password'], con['host'], destination, source)
+                        self.download_script+='echo -n download "%s"\n' % (destination, )
+                        self.download_script+='ftpget -P ${PORT} -u "${USERNAME}" -p "${PASSWORD}" "${HOST}" "${DOWNLOAD_DIR}/%s" "%s" && echo " OK" || echo " ERR"\n' % (relfilename, destination, )
+                        self.log.info('start upload of "%s" to "%s"', source, destination)
+                        cmd_line=self.hide_password(ftp_line)
+                        self.log.debug('ftp command: %s', cmd_line)
+                        exit_code, output=RunCommand(self.t, ftp_line, log=None)
+                        if exit_code==0:
+                            self.log.debug('file uploaded: %s', destination)
                         else:
-                            destination=os.path.join(target, relfilename)
-                        self.upload_queue_add(protocol, con, vm, filename, destination)
+                            status='ERR'
+                            self.log.error('upload failed %s (%s)', destination, output)
+                    elif protocol=='ssh':
+                        dest_dir=os.path.join(target, posixpath.dirname(destination))
+                        try:
+                            if not os.path.isdir(dest_dir):
+                                self.log.debug('create directory %s', dest_dir)
+                                os.makedirs(dest_dir)
+                        except Exception, e:
+                            self.log.error('creating directory %s: %s', target, e)
+                            return False
+                        
+                        self.log.info('start download from "%s" to "%s"', source, destination)
+        
+                        # escape space characters
+                        scp_source=source.replace(' ', '\\ ')
+                        if not self.scp_bin:
+                            # use paramiko
+                            try:
+                                scpclient=MySCPClient(self.t, buff_size=65536)
+                                self.log.debug('scp %s %s', scp_source, destination)
+                                scpclient.get(source, destination)
+                            except scp.SCPException, e:
+                                self.log.error('download failed: %s (%s)', destination, e)
+                                status='ERR'
+                        else:
+                            # use local SCP
+                            scp_args=self.scp_args[:]
+                            scp_args+=[ '%s@%s:%s' % (self.login, self.host, scp_source), destination]
+                            scp_cmd=self.hide_password(' '.join(scp_args))
+                            self.log.debug('scp_cmd=%s', scp_cmd)
+                            if sys.platform in ('win32', ):
+                                scp_args=param_encode(scp_args, self.manager.default_encoding)
+                            scp_process=subprocess.Popen(scp_args, executable=self.scp_bin, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+                            scp_out, scp_err=scp_process.communicate()
+                            if scp_out:
+                                self.log.debug('copying %s: %s', source, scp_out)
+                            if scp_err:
+                                self.log.debug('copying %s: %s', source, scp_err)
+        
+                            if scp_process.returncode!=0:
+                                self.log.error('download failed: %s (%s)', destination, scp_out+scp_err)
+                                status='ERR'
+
+            end=time.time()                
+            if status!='OK':
+                self.upload_status_code='ERR'
+            self.upload_status+='upload_status-%s=%s\r\n' % (current_vm, status)
+            self.upload_status+='upload_end-%s=%s\r\n' % (current_vm, time.ctime(end))
+            self.upload_status+='upload_end_epoch-%s=%d\r\n' % (current_vm, end)
+            
+            if target and type=='move':
+                # remove source
+                exit_code, output=RunCommand(self.t, 'rm -rf "%s"' % (find_src, ), log=None)
+                if exit_code:
+                    self.log.error('failed to remove %s: %r', find_src, output)
+                    self.upload_status_code='ERR'
+                else:
+                    self.log.info('remove %s', find_src)
+
+                            
+        self.upload_status='upload_status=%s\r\n%s' % (self.upload_status_code, self.upload_status)
+        return True
 
     # -----------------------------------------------------------------------
     def run(self, command, job, manager):
@@ -656,10 +849,10 @@ class Ghettovcb(Archiver):
         chan.exec_command(self.cmd_line)
 
         attachments=[]
-        
+        self.finished_vm=set()
         output=''
         state='waiting_for_start'
-        current_vmdk=None
+        current_vm=None
         for line in socket_readlines(chan):
             # I'm reading the socket in this loop,
             # I CANNOT DO SOMETHING TIME CONSUMING or do it in background 
@@ -671,6 +864,11 @@ class Ghettovcb(Archiver):
                 vm_backup_volume=match.group('vm_backup_volume')
                 continue
 
+            match=self.config_disk_backup_format_re.match(line)
+            if match:
+                disk_backup_format=match.group('disk_backup_format')
+                continue
+
             match=self.config_vm_backup_dir_naming_convention_re.match(line)
             if match:
                 vm_backup_dir_naming_convention=match.group('vm_backup_dir_naming_convention')
@@ -694,32 +892,33 @@ class Ghettovcb(Archiver):
                 # ENABLE_NON_PERSISTENT_NFS
                 target=posixpath.join(vm_backup_volume, current_vm)
                 compress=vm_var['ENABLE_COMPRESSION']=='1'
-
+                persistent_nfs=vm_var['ENABLE_NON_PERSISTENT_NFS']!='1'
+                root_backup_dir=posixpath.join(vm_backup_volume, current_vm, current_vm+'-'+vm_backup_dir_naming_convention)
+                # I cannot start upload_vm now, because of a race condition, I need to wait for
+                # the start of first disk to besure direcrories exists
+                wait_first_disk=True
                 continue
 
             match=self.find_vmdk_re.match(line)
             if match:
                 # start a new disk
-                if current_vmdk and vm_var['ENABLE_NON_PERSISTENT_NFS']!='1' and self.destination:
-                    # if needed, now we can upload the previous disk
-                    _res=self.upload_vmdk(current_vm, root_backup_dir, current_vmdk, now)
-                     
-                # store current vmdk
-                current_vmdk=match.group('vmdk')
-                root_backup_dir=posixpath.join(vm_backup_volume, current_vm, current_vm+'-'+vm_backup_dir_naming_convention)
+                if wait_first_disk and self.destination and not compress and persistent_nfs:
+                    # monitor directory and an upload "finished file" file 
+                    self.upload_vm(current_vm, root_backup_dir, now, False)
+                wait_first_disk=False
                 continue
 
             match=self.backup_duration_re.match(line)
             if match:
-                """the backup just finished"""
-                if current_vmdk and vm_var['ENABLE_NON_PERSISTENT_NFS']!='1' and self.destination:
-                    # if needed, now we can upload the previous disk
-                    _res=self.upload_vmdk(current_vm, root_backup_dir, current_vmdk, now)
+                # the backup just finished
+                self.finished_vm.add(current_vm)
+                if self.destination and compress and persistent_nfs:
+                    # upload compressed file 
+                    self.upload_vm(current_vm, posixpath.dirname(root_backup_dir), now, posixpath.basename(root_backup_dir)+'.gz')
             
                 # search the target and make a directory listing 
-                
                 dir_lst, up_lst, scp_process=None, None, None
-                if vm_var['ENABLE_NON_PERSISTENT_NFS']=='1' and not self.ghettovcb_wait_nfs:
+                if not persistent_nfs and not self.ghettovcb_wait_nfs:
                     log.warning('cannot retrieve data from non persistent NFS share on old version of ghettoVCB.sh')
                 else:
                     if compress:
@@ -747,7 +946,6 @@ class Ghettovcb(Archiver):
             
                     if up_lst:
                         attachments.append(('dirup-%s.txt' % (current_vm,), None, up_lst, 'text', 'plain', 'utf-8'))
-            
         
         ghettovcb_exit_code=chan.recv_exit_status()
         chan.close()
@@ -770,12 +968,13 @@ class Ghettovcb(Archiver):
         ls_volumes='== directory /vmfs/volumes\n'+ls_volumes
 
         # Nothing more to upload    
-        self.upload_queue_add('done', None, None, None, None)
-        # wait for end of upload /download
-        if not self.upload_queue.empty():
-            log.info('waiting end of upload...')
-            while not self.upload_queue.empty():
-                time.sleep(5)
+        if self.upload_thread:
+            self.upload_vm(None, None ,None, None)
+            # wait for end of upload/download
+            log.info('waiting for end of upload...')
+            self.upload_thread.join()
+            if self.download_script:
+                attachments.insert(0, ('download.txt', None, self.download_script, 'text', 'plain', 'utf-8') )
 
         end=int(time.time())
 
